@@ -16,13 +16,15 @@
 module Data.Primitive.Array (
   Array(..), MutableArray(..),
 
-  newArray, readArray, writeArray, indexArray, indexArrayM,
-  freezeArray, thawArray,
+  newArray, readArray, writeArray, indexArray, indexArrayM, indexArray##,
+  freezeArray, thawArray, runArray,
   unsafeFreezeArray, unsafeThawArray, sameMutableArray,
   copyArray, copyMutableArray,
   cloneArray, cloneMutableArray,
   sizeofArray, sizeofMutableArray,
-  fromListN, fromList
+  fromListN, fromList,
+  mapArray',
+  traverseArrayP
 ) where
 
 import Control.Monad.Primitive
@@ -42,7 +44,7 @@ import Data.Primitive.Internal.Compat ( isTrue#, mkNoRepType )
 import Control.Monad.ST(ST,runST)
 
 import Control.Applicative
-import Control.Monad (MonadPlus(..))
+import Control.Monad (MonadPlus(..), when)
 import Control.Monad.Fix
 #if MIN_VERSION_base(4,4,0)
 import Control.Monad.Zip
@@ -53,31 +55,35 @@ import Data.Traversable (Traversable(..))
 import Data.Monoid
 #endif
 #if MIN_VERSION_base(4,9,0)
+import qualified GHC.ST as GHCST
 import qualified Data.Foldable as F
 import Data.Semigroup
 #endif
+#if MIN_VERSION_base(4,8,0)
+import Data.Functor.Identity
+#endif
+#if MIN_VERSION_base(4,10,0)
+import GHC.Exts (runRW#)
+#elif MIN_VERSION_base(4,9,0)
+import GHC.Base (runRW#)
+#endif
 
 import Text.ParserCombinators.ReadP
 
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,4,0)
+import Data.Functor.Classes (Eq1(..),Ord1(..),Show1(..),Read1(..))
+#endif
+
 -- | Boxed arrays
 data Array a = Array
-             { array# :: Array# a
-#if (__GLASGOW_HASKELL__ < 702)
-             , sizeofArray :: {-# UNPACK #-} !Int
-#endif
-             }
+  { array# :: Array# a }
   deriving ( Typeable )
 
 -- | Mutable boxed arrays associated with a primitive state token.
 data MutableArray s a = MutableArray
-                      { marray# :: MutableArray# s a
-#if (__GLASGOW_HASKELL__ < 702)
-                      , sizeofMutableArray :: {-# UNPACK #-} !Int
-#endif
-                      }
+  { marray# :: MutableArray# s a }
   deriving ( Typeable )
 
-#if (__GLASGOW_HASKELL__ >= 702)
 sizeofArray :: Array a -> Int
 sizeofArray a = I# (sizeofArray# (array# a))
 {-# INLINE sizeofArray #-}
@@ -85,7 +91,6 @@ sizeofArray a = I# (sizeofArray# (array# a))
 sizeofMutableArray :: MutableArray s a -> Int
 sizeofMutableArray a = I# (sizeofMutableArray# (marray# a))
 {-# INLINE sizeofMutableArray #-}
-#endif
 
 -- | Create a new mutable array of the specified size and initialise all
 -- elements with the given value.
@@ -95,9 +100,6 @@ newArray (I# n#) x = primitive
    (\s# -> case newArray# n# x s# of
              (# s'#, arr# #) ->
                let ma = MutableArray arr#
-#if (__GLASGOW_HASKELL__ < 702)
-                          (I# n#)
-#endif
                in (# s'# , ma #))
 
 -- | Read a value from the array at the given index.
@@ -115,6 +117,13 @@ indexArray :: Array a -> Int -> a
 {-# INLINE indexArray #-}
 indexArray arr (I# i#) = case indexArray# (array# arr) i# of (# x #) -> x
 
+-- | Read a value from the immutable array at the given index, returning
+-- the result in an unboxed unary tuple. This is currently used to implement
+-- folds.
+indexArray## :: Array a -> Int -> (# a #)
+indexArray## arr (I# i) = indexArray# (array# arr) i
+{-# INLINE indexArray## #-}
+
 -- | Monadically read a value from the immutable array at the given index.
 -- This allows us to be strict in the array while remaining lazy in the read
 -- element which is very useful for collective operations. Suppose we want to
@@ -154,16 +163,9 @@ freezeArray
   -> Int                          -- ^ length
   -> m (Array a)
 {-# INLINE freezeArray #-}
-#if (__GLASGOW_HASKELL__ >= 702)
 freezeArray (MutableArray ma#) (I# off#) (I# len#) =
   primitive $ \s -> case freezeArray# ma# off# len# s of
     (# s', a# #) -> (# s', Array a# #)
-#else
-freezeArray src off len = do
-  dst <- newArray len (die "freezeArray" "impossible")
-  copyMutableArray dst 0 src off len
-  unsafeFreezeArray dst
-#endif
 
 -- | Convert a mutable array to an immutable one without copying. The
 -- array should not be modified after the conversion.
@@ -173,9 +175,6 @@ unsafeFreezeArray arr
   = primitive (\s# -> case unsafeFreezeArray# (marray# arr) s# of
                         (# s'#, arr'# #) ->
                           let a = Array arr'#
-#if (__GLASGOW_HASKELL__ < 702)
-                                    (sizeofMutableArray arr)
-#endif
                           in (# s'#, a #))
 
 -- | Create a mutable array from a slice of an immutable array.
@@ -189,16 +188,9 @@ thawArray
   -> Int     -- ^ length
   -> m (MutableArray (PrimState m) a)
 {-# INLINE thawArray #-}
-#if (__GLASGOW_HASKELL__ >= 702)
 thawArray (Array a#) (I# off#) (I# len#) =
   primitive $ \s -> case thawArray# a# off# len# s of
     (# s', ma# #) -> (# s', MutableArray ma# #)
-#else
-thawArray src off len = do
-  dst <- newArray len (die "thawArray" "impossible")
-  copyArray dst 0 src off len
-  return dst
-#endif
 
 -- | Convert an immutable array to an mutable one without copying. The
 -- immutable array should not be used after the conversion.
@@ -208,9 +200,6 @@ unsafeThawArray a
   = primitive (\s# -> case unsafeThawArray# (array# a) s# of
                         (# s'#, arr'# #) ->
                           let ma = MutableArray arr'#
-#if (__GLASGOW_HASKELL__ < 702)
-                                     (sizeofArray a)
-#endif
                           in (# s'#, ma #))
 
 -- | Check whether the two arrays refer to the same memory block.
@@ -275,15 +264,8 @@ cloneArray :: Array a -- ^ source array
            -> Int     -- ^ number of elements to copy
            -> Array a
 {-# INLINE cloneArray #-}
-#if __GLASGOW_HASKELL__ >= 702
 cloneArray (Array arr#) (I# off#) (I# len#)
   = case cloneArray# arr# off# len# of arr'# -> Array arr'#
-#else
-cloneArray arr off len = runST $ do
-    marr2 <- newArray len $ die "cloneArray" "impossible"
-    copyArray marr2 0 arr off len
-    unsafeFreezeArray marr2
-#endif
 
 -- | Return a newly allocated MutableArray. with the specified subrange of
 -- the provided MutableArray. The provided MutableArray should contain the
@@ -294,112 +276,209 @@ cloneMutableArray :: PrimMonad m
         -> Int                          -- ^ number of elements to copy
         -> m (MutableArray (PrimState m) a)
 {-# INLINE cloneMutableArray #-}
-#if __GLASGOW_HASKELL__ >= 702
 cloneMutableArray (MutableArray arr#) (I# off#) (I# len#) = primitive
    (\s# -> case cloneMutableArray# arr# off# len# s# of
              (# s'#, arr'# #) -> (# s'#, MutableArray arr'# #))
-#else
-cloneMutableArray marr off len = do
-        marr2 <- newArray len $ die "cloneMutableArray" "impossible"
-        let go !i !j c
-                | c >= len = return marr2
-                | otherwise = do
-                    b <- readArray marr i
-                    writeArray marr2 j b
-                    go (i+1) (j+1) (c+1)
-        go off 0 0
-#endif
 
 emptyArray :: Array a
 emptyArray =
   runST $ newArray 0 (die "emptyArray" "impossible") >>= unsafeFreezeArray
 {-# NOINLINE emptyArray #-}
 
+#if !MIN_VERSION_base(4,9,0)
 createArray
   :: Int
   -> a
   -> (forall s. MutableArray s a -> ST s ())
   -> Array a
 createArray 0 _ _ = emptyArray
-createArray n x f = runST $ do
-  ma <- newArray n x
-  f ma
-  unsafeFreezeArray ma
+createArray n x f = runArray $ do
+  mary <- newArray n x
+  f mary
+  pure mary
+
+runArray
+  :: (forall s. ST s (MutableArray s a))
+  -> Array a
+runArray m = runST $ m >>= unsafeFreezeArray
+
+#else /* Below, runRW# is available. */
+
+-- This low-level business is designed to work with GHC's worker-wrapper
+-- transformation. A lot of the time, we don't actually need an Array
+-- constructor. By putting it on the outside, and being careful about
+-- how we special-case the empty array, we can make GHC smarter about this.
+-- The only downside is that separately created 0-length arrays won't share
+-- their Array constructors, although they'll share their underlying
+-- Array#s.
+createArray
+  :: Int
+  -> a
+  -> (forall s. MutableArray s a -> ST s ())
+  -> Array a
+createArray 0 _ _ = Array (emptyArray# (# #))
+createArray n x f = runArray $ do
+  mary <- newArray n x
+  f mary
+  pure mary
+
+runArray
+  :: (forall s. ST s (MutableArray s a))
+  -> Array a
+runArray m = Array (runArray# m)
+
+runArray#
+  :: (forall s. ST s (MutableArray s a))
+  -> Array# a
+runArray# m = case runRW# $ \s ->
+  case unST m s of { (# s', MutableArray mary# #) ->
+  unsafeFreezeArray# mary# s'} of (# _, ary# #) -> ary#
+
+unST :: ST s a -> State# s -> (# State# s, a #)
+unST (GHCST.ST f) = f
+
+emptyArray# :: (# #) -> Array# a
+emptyArray# _ = case emptyArray of Array ar -> ar
+{-# NOINLINE emptyArray# #-}
+#endif
+
 
 die :: String -> String -> a
 die fun problem = error $ "Data.Primitive.Array." ++ fun ++ ": " ++ problem
 
+arrayLiftEq :: (a -> b -> Bool) -> Array a -> Array b -> Bool
+arrayLiftEq p a1 a2 = sizeofArray a1 == sizeofArray a2 && loop (sizeofArray a1 - 1)
+  where loop i | i < 0     = True
+               | (# x1 #) <- indexArray## a1 i
+               , (# x2 #) <- indexArray## a2 i
+               , otherwise = p x1 x2 && loop (i-1)
+
 instance Eq a => Eq (Array a) where
-  a1 == a2 = sizeofArray a1 == sizeofArray a2 && loop (sizeofArray a1 - 1)
-   where loop i | i < 0     = True
-                | otherwise = indexArray a1 i == indexArray a2 i && loop (i-1)
+  a1 == a2 = arrayLiftEq (==) a1 a2
+
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,4,0)
+-- | @since 0.6.4.0
+instance Eq1 Array where
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,5,0)
+  liftEq = arrayLiftEq
+#else
+  eq1 = arrayLiftEq (==)
+#endif
+#endif
 
 instance Eq (MutableArray s a) where
   ma1 == ma2 = isTrue# (sameMutableArray# (marray# ma1) (marray# ma2))
 
+arrayLiftCompare :: (a -> b -> Ordering) -> Array a -> Array b -> Ordering
+arrayLiftCompare elemCompare a1 a2 = loop 0
+  where
+  mn = sizeofArray a1 `min` sizeofArray a2
+  loop i
+    | i < mn
+    , (# x1 #) <- indexArray## a1 i
+    , (# x2 #) <- indexArray## a2 i
+    = elemCompare x1 x2 `mappend` loop (i+1)
+    | otherwise = compare (sizeofArray a1) (sizeofArray a2)
+
+-- | Lexicographic ordering. Subject to change between major versions.
 instance Ord a => Ord (Array a) where
-  compare a1 a2 = loop 0
-   where
-   mn = sizeofArray a1 `min` sizeofArray a2
-   loop i
-     | i < mn    = compare (indexArray a1 i) (indexArray a2 i) `mappend` loop (i+1)
-     | otherwise = compare (sizeofArray a1) (sizeofArray a2)
+  compare a1 a2 = arrayLiftCompare compare a1 a2
+
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,4,0)
+-- | @since 0.6.4.0
+instance Ord1 Array where
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,5,0)
+  liftCompare = arrayLiftCompare
+#else
+  compare1 = arrayLiftCompare compare
+#endif
+#endif
 
 instance Foldable Array where
-  foldr f z a = go 0
-   where go i | i < sizeofArray a = f (indexArray a i) (go $ i+1)
-              | otherwise         = z
+  -- Note: we perform the array lookups eagerly so we won't
+  -- create thunks to perform lookups even if GHC can't see
+  -- that the folding function is strict.
+  foldr f = \z !ary ->
+    let
+      !sz = sizeofArray ary
+      go i
+        | i == sz = z
+        | (# x #) <- indexArray## ary i
+        = f x (go (i+1))
+    in go 0
   {-# INLINE foldr #-}
-  foldl f z a = go (sizeofArray a - 1)
-   where go i | i < 0     = z
-              | otherwise = f (go $ i-1) (indexArray a i)
+  foldl f = \z !ary ->
+    let
+      go i
+        | i < 0 = z
+        | (# x #) <- indexArray## ary i
+        = f (go (i-1)) x
+    in go (sizeofArray ary - 1)
   {-# INLINE foldl #-}
-  foldr1 f a | sz < 0    = die "foldr1" "empty array"
-             | otherwise = go 0
-   where sz = sizeofArray a - 1
-         z = indexArray a sz
-         go i | i < sz    = f (indexArray a i) (go $ i+1)
-              | otherwise = z
+  foldr1 f = \ !ary ->
+    let
+      !sz = sizeofArray ary - 1
+      go i =
+        case indexArray## ary i of
+          (# x #) | i == sz -> x
+                  | otherwise -> f x (go (i+1))
+    in if sz < 0
+       then die "foldr1" "empty array"
+       else go 0
   {-# INLINE foldr1 #-}
-  foldl1 f a | sz == 0   = die "foldl1" "empty array"
-             | otherwise = go $ sz-1
-   where sz = sizeofArray a
-         z = indexArray a 0
-         go i | i < 1     = f (go $ i-1) (indexArray a i)
-              | otherwise = z
+  foldl1 f = \ !ary ->
+    let
+      !sz = sizeofArray ary - 1
+      go i =
+        case indexArray## ary i of
+          (# x #) | i == 0 -> x
+                  | otherwise -> f (go (i - 1)) x
+    in if sz < 0
+       then die "foldl1" "empty array"
+       else go sz
   {-# INLINE foldl1 #-}
 #if MIN_VERSION_base(4,6,0)
-  foldr' f z a = go (sizeofArray a - 1) z
-   where go i !acc | i < 0     = acc
-                   | otherwise = go (i-1) (f (indexArray a i) acc)
+  foldr' f = \z !ary ->
+    let
+      go i !acc
+        | i == -1 = acc
+        | (# x #) <- indexArray## ary i
+        = go (i-1) (f x acc)
+    in go (sizeofArray ary - 1) z
   {-# INLINE foldr' #-}
-  foldl' f z a = go 0 z
-   where go i !acc | i < sizeofArray a = go (i+1) (f acc $ indexArray a i)
-                   | otherwise         = acc
+  foldl' f = \z !ary ->
+    let
+      !sz = sizeofArray ary
+      go i !acc
+        | i == sz = acc
+        | (# x #) <- indexArray## ary i
+        = go (i+1) (f acc x)
+    in go 0 z
   {-# INLINE foldl' #-}
 #endif
 #if MIN_VERSION_base(4,8,0)
-  toList a = Exts.build $ \c z -> let
-      sz = sizeofArray a
-      go i | i < sz    = c (indexArray a i) (go $ i+1)
-           | otherwise = z
-    in go 0
-  {-# INLINE toList #-}
   null a = sizeofArray a == 0
   {-# INLINE null #-}
   length = sizeofArray
   {-# INLINE length #-}
-  maximum a | sz == 0   = die "maximum" "empty array"
-            | otherwise = go 1 (indexArray a 0)
-   where sz = sizeofArray a
-         go i !e | i < sz    = go (i+1) (max e $ indexArray a i)
-                 | otherwise = e
+  maximum ary | sz == 0   = die "maximum" "empty array"
+              | (# frst #) <- indexArray## ary 0
+              = go 1 frst
+   where
+     sz = sizeofArray ary
+     go i !e
+       | i == sz = e
+       | (# x #) <- indexArray## ary i
+       = go (i+1) (max e x)
   {-# INLINE maximum #-}
-  minimum a | sz == 0   = die "minimum" "empty array"
-            | otherwise = go 1 (indexArray a 0)
-   where sz = sizeofArray a
-         go i !e | i < sz    = go (i+1) (min e $ indexArray a i)
-                 | otherwise = e
+  minimum ary | sz == 0   = die "minimum" "empty array"
+              | (# frst #) <- indexArray## ary 0
+              = go 1 frst
+   where sz = sizeofArray ary
+         go i !e
+           | i == sz = e
+           | (# x #) <- indexArray## ary i
+           = go (i+1) (min e x)
   {-# INLINE minimum #-}
   sum = foldl' (+) 0
   {-# INLINE sum #-}
@@ -407,56 +486,155 @@ instance Foldable Array where
   {-# INLINE product #-}
 #endif
 
+newtype STA a = STA {_runSTA :: forall s. MutableArray# s a -> ST s (Array a)}
+
+runSTA :: Int -> STA a -> Array a
+runSTA !sz = \ (STA m) -> runST $ newArray_ sz >>= \ ar -> m (marray# ar)
+{-# INLINE runSTA #-}
+
+newArray_ :: Int -> ST s (MutableArray s a)
+newArray_ !n = newArray n badTraverseValue
+
+badTraverseValue :: a
+badTraverseValue = die "traverse" "bad indexing"
+{-# NOINLINE badTraverseValue #-}
+
 instance Traversable Array where
-  traverse f a =
-    fromListN (sizeofArray a)
-      <$> traverse (f . indexArray a) [0 .. sizeofArray a - 1]
+  traverse f = traverseArray f
+  {-# INLINE traverse #-}
+
+traverseArray
+  :: Applicative f
+  => (a -> f b)
+  -> Array a
+  -> f (Array b)
+traverseArray f = \ !ary ->
+  let
+    !len = sizeofArray ary
+    go !i
+      | i == len = pure $ STA $ \mary -> unsafeFreezeArray (MutableArray mary)
+      | (# x #) <- indexArray## ary i
+      = liftA2 (\b (STA m) -> STA $ \mary ->
+                  writeArray (MutableArray mary) i b >> m mary)
+               (f x) (go (i + 1))
+  in if len == 0
+     then pure emptyArray
+     else runSTA len <$> go 0
+{-# INLINE [1] traverseArray #-}
+
+{-# RULES
+"traverse/ST" forall (f :: a -> ST s b). traverseArray f =
+   traverseArrayP f
+"traverse/IO" forall (f :: a -> IO b). traverseArray f =
+   traverseArrayP f
+ #-}
+#if MIN_VERSION_base(4,8,0)
+{-# RULES
+"traverse/Id" forall (f :: a -> Identity b). traverseArray f =
+   (coerce :: (Array a -> Array (Identity b))
+           -> Array a -> Identity (Array b)) (fmap f)
+ #-}
+#endif
+
+-- | This is the fastest, most straightforward way to traverse
+-- an array, but it only works correctly with a sufficiently
+-- "affine" 'PrimMonad' instance. In particular, it must only produce
+-- *one* result array. 'Control.Monad.Trans.List.ListT'-transformed
+-- monads, for example, will not work right at all.
+traverseArrayP
+  :: PrimMonad m
+  => (a -> m b)
+  -> Array a
+  -> m (Array b)
+traverseArrayP f = \ !ary ->
+  let
+    !sz = sizeofArray ary
+    go !i !mary
+      | i == sz
+      = unsafeFreezeArray mary
+      | otherwise
+      = do
+          a <- indexArrayM ary i
+          b <- f a
+          writeArray mary i b
+          go (i + 1) mary
+  in do
+    mary <- newArray sz badTraverseValue
+    go 0 mary
+{-# INLINE traverseArrayP #-}
+
+-- | Strict map over the elements of the array.
+mapArray' :: (a -> b) -> Array a -> Array b
+mapArray' f a =
+  createArray (sizeofArray a) (die "mapArray'" "impossible") $ \mb ->
+    let go i | i == sizeofArray a
+             = return ()
+             | otherwise
+             = do x <- indexArrayM a i
+                  -- We use indexArrayM here so that we will perform the
+                  -- indexing eagerly even if f is lazy.
+                  let !y = f x
+                  writeArray mb i y >> go (i+1)
+     in go 0
+{-# INLINE mapArray' #-}
+
+arrayFromListN :: Int -> [a] -> Array a
+arrayFromListN n l =
+  createArray n (die "fromListN" "uninitialized element") $ \sma ->
+    let go !ix [] = if ix == n
+          then return ()
+          else die "fromListN" "list length less than specified size"
+        go !ix (x : xs) = if ix < n
+          then do
+            writeArray sma ix x
+            go (ix+1) xs
+          else die "fromListN" "list length greater than specified size"
+    in go 0 l
+
+arrayFromList :: [a] -> Array a
+arrayFromList l = arrayFromListN (length l) l
 
 #if MIN_VERSION_base(4,7,0)
 instance Exts.IsList (Array a) where
   type Item (Array a) = a
-  fromListN n l =
-    createArray n (die "fromListN" "mismatched size and list") $ \mi ->
-      let go i (x:xs) = writeArray mi i x >> go (i+1) xs
-          go _ [    ] = return ()
-       in go 0 l
-  fromList l = Exts.fromListN (length l) l
+  fromListN = arrayFromListN
+  fromList = arrayFromList
   toList = toList
 #else
 fromListN :: Int -> [a] -> Array a
-fromListN n l =
-  createArray n (die "fromListN" "mismatched size and list") $ \mi ->
-    let go i (x:xs) = writeArray mi i x >> go (i+1) xs
-        go _ [    ] = return ()
-     in go 0 l
+fromListN = arrayFromListN
 
 fromList :: [a] -> Array a
-fromList l = fromListN (length l) l
+fromList = arrayFromList
 #endif
 
 instance Functor Array where
   fmap f a =
     createArray (sizeofArray a) (die "fmap" "impossible") $ \mb ->
-      let go i | i < sizeofArray a = return ()
-               | otherwise         = writeArray mb i (f $ indexArray a i)
-                                  >> go (i+1)
+      let go i | i == sizeofArray a
+               = return ()
+               | otherwise
+               = do x <- indexArrayM a i
+                    writeArray mb i (f x) >> go (i+1)
        in go 0
 #if MIN_VERSION_base(4,8,0)
-  e <$ a = runST $ newArray (sizeofArray a) e >>= unsafeFreezeArray
+  e <$ a = createArray (sizeofArray a) e (\ !_ -> pure ())
 #endif
 
 instance Applicative Array where
-  pure x = runST $ newArray 1 x >>= unsafeFreezeArray
-  ab <*> a = runST $ do
-    mb <- newArray (szab*sza) $ die "<*>" "impossible"
-    let go1 i
-          | i < szab  = go2 (i*sza) (indexArray ab i) 0 >> go1 (i+1)
-          | otherwise = return ()
-        go2 off f j
-          | j < sza   = writeArray mb (off + j) (f $ indexArray a j)
-          | otherwise = return ()
-    go1 0
-    unsafeFreezeArray mb
+  pure x = runArray $ newArray 1 x
+  ab <*> a = createArray (szab*sza) (die "<*>" "impossible") $ \mb ->
+    let go1 i = when (i < szab) $
+            do
+              f <- indexArrayM ab i
+              go2 (i*sza) f 0
+              go1 (i+1)
+        go2 off f j = when (j < sza) $
+            do
+              x <- indexArrayM a j
+              writeArray mb (off + j) (f x)
+              go2 off f (j + 1)
+    in go1 0
    where szab = sizeofArray ab ; sza = sizeofArray a
   a *> b = createArray (sza*szb) (die "*>" "impossible") $ \mb ->
     let go i | i < sza   = copyArray mb (i * szb) b 0 szb
@@ -466,7 +644,9 @@ instance Applicative Array where
   a <* b = createArray (sza*szb) (die "<*" "impossible") $ \ma ->
     let fill off i e | i < szb   = writeArray ma (off+i) e >> fill off (i+1) e
                      | otherwise = return ()
-        go i | i < sza   = fill (i*szb) 0 (indexArray a i) >> go (i+1)
+        go i | i < sza
+             = do x <- indexArrayM a i
+                  fill (i*szb) 0 x >> go (i+1)
              | otherwise = return ()
      in go 0
    where sza = sizeofArray a ; szb = sizeofArray b
@@ -481,20 +661,36 @@ instance Alternative Array where
   many a | sizeofArray a == 0 = pure []
          | otherwise = die "many" "infinite arrays are not well defined"
 
+data ArrayStack a
+  = PushArray !(Array a) !(ArrayStack a)
+  | EmptyStack
+-- See the note in SmallArray about how we might improve this.
+
 instance Monad Array where
   return = pure
   (>>) = (*>)
-  a >>= f = push 0 [] (sizeofArray a - 1)
-   where
-   push !sz bs i
-     | i < 0 = build sz bs
-     | otherwise = let b = f $ indexArray a i
-                    in push (sz + sizeofArray b) (b:bs) (i+1)
 
-   build sz stk = createArray sz (die ">>=" "impossible") $ \mb ->
-     let go off (b:bs) = copyArray mb off b 0 (sizeofArray b) >> go (off + sizeofArray b) bs
-         go _   [    ] = return ()
-      in go 0 stk
+  ary >>= f = collect 0 EmptyStack (la-1)
+   where
+   la = sizeofArray ary
+   collect sz stk i
+     | i < 0 = createArray sz (die ">>=" "impossible") $ fill 0 stk
+     | (# x #) <- indexArray## ary i
+     , let sb = f x
+           lsb = sizeofArray sb
+       -- If we don't perform this check, we could end up allocating
+       -- a stack full of empty arrays if someone is filtering most
+       -- things out. So we refrain from pushing empty arrays.
+     = if lsb == 0
+       then collect sz stk (i - 1)
+       else collect (sz + lsb) (PushArray sb stk) (i-1)
+
+   fill _   EmptyStack         _   = return ()
+   fill off (PushArray sb sbs) smb
+     | let lsb = sizeofArray sb
+     = copyArray smb off sb 0 (lsb)
+         *> fill (off + lsb) sbs smb
+
   fail _ = empty
 
 instance MonadPlus Array where
@@ -503,10 +699,13 @@ instance MonadPlus Array where
 
 zipW :: String -> (a -> b -> c) -> Array a -> Array b -> Array c
 zipW s f aa ab = createArray mn (die s "impossible") $ \mc ->
-  let go i
-        | i < mn    = writeArray mc i (f (indexArray aa i) (indexArray ab i))
-                   >> go (i+1)
-        | otherwise = return ()
+  let go i | i < mn
+           = do
+               x <- indexArrayM aa i
+               y <- indexArrayM ab i
+               writeArray mc i (f x y)
+               go (i+1)
+           | otherwise = return ()
    in go 0
  where mn = sizeofArray aa `min` sizeofArray ab
 {-# INLINE zipW #-}
@@ -520,7 +719,7 @@ instance MonadZip Array where
     ma <- newArray sz (die "munzip" "impossible")
     mb <- newArray sz (die "munzip" "impossible")
     let go i | i < sz = do
-          let (a, b) = indexArray aab i
+          (a, b) <- indexArrayM aab i
           writeArray ma i a
           writeArray mb i b
           go (i+1)
@@ -530,9 +729,17 @@ instance MonadZip Array where
 #endif
 
 instance MonadFix Array where
-  mfix f = let l = mfix (toList . f) in fromListN (length l) l
+  mfix f = createArray (sizeofArray (f err))
+                       (die "mfix" "impossible") $ flip fix 0 $
+    \r !i !mary -> when (i < sz) $ do
+                      writeArray mary i (fix (\xi -> f xi `indexArray` i))
+                      r (i + 1) mary
+    where
+      sz = sizeofArray (f err)
+      err = error "mfix for Data.Primitive.Array applied to strict function."
 
 #if MIN_VERSION_base(4,9,0)
+-- | @since 0.6.3.0
 instance Semigroup (Array a) where
   (<>) = (<|>)
   sconcat = mconcat . F.toList
@@ -550,19 +757,50 @@ instance Monoid (Array a) where
      in go 0 l
    where sz = sum . fmap sizeofArray $ l
 
+arrayLiftShowsPrec :: (Int -> a -> ShowS) -> ([a] -> ShowS) -> Int -> Array a -> ShowS
+arrayLiftShowsPrec elemShowsPrec elemListShowsPrec p a = showParen (p > 10) $
+  showString "fromListN " . shows (sizeofArray a) . showString " "
+    . listLiftShowsPrec elemShowsPrec elemListShowsPrec 11 (toList a)
+
+-- this need to be included for older ghcs
+listLiftShowsPrec :: (Int -> a -> ShowS) -> ([a] -> ShowS) -> Int -> [a] -> ShowS
+listLiftShowsPrec _ sl _ = sl
+
 instance Show a => Show (Array a) where
-  showsPrec p a = showParen (p > 10) $
-    showString "fromListN " . shows (sizeofArray a) . showString " "
-      . shows (toList a)
+  showsPrec p a = arrayLiftShowsPrec showsPrec showList p a
+
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,4,0)
+-- | @since 0.6.4.0
+instance Show1 Array where
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,5,0)
+  liftShowsPrec = arrayLiftShowsPrec
+#else
+  showsPrec1 = arrayLiftShowsPrec showsPrec showList
+#endif
+#endif
+
+arrayLiftReadsPrec :: (Int -> ReadS a) -> ReadS [a] -> Int -> ReadS (Array a)
+arrayLiftReadsPrec _ listReadsPrec p = readParen (p > 10) . readP_to_S $ do
+  () <$ string "fromListN"
+  skipSpaces
+  n <- readS_to_P reads
+  skipSpaces
+  l <- readS_to_P listReadsPrec
+  return $ arrayFromListN n l
 
 instance Read a => Read (Array a) where
-  readsPrec p = readParen (p > 10) . readP_to_S $ do
-    () <$ string "fromListN"
-    skipSpaces
-    n <- readS_to_P reads
-    skipSpaces
-    l <- readS_to_P reads
-    return $ fromListN n l
+  readsPrec = arrayLiftReadsPrec readsPrec readList
+
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,4,0)
+-- | @since 0.6.4.0
+instance Read1 Array where
+#if MIN_VERSION_base(4,9,0) || MIN_VERSION_transformers(0,5,0)
+  liftReadsPrec = arrayLiftReadsPrec
+#else
+  readsPrec1 = arrayLiftReadsPrec readsPrec readList
+#endif
+#endif
+
 
 arrayDataType :: DataType
 arrayDataType = mkDataType "Data.Primitive.Array.Array" [fromListConstr]
